/*--------------------------------------------------------------------
 * output_viewer.h
 *
 * The output_viewer displays C output generated by s2c.
 * There should be `line' annotations embedded as comments in the C
 * file, so that the viewer can map the output lines back to Suif.
 *
 */

#ifndef OUTPUT_VIEWER_H
#define OUTPUT_VIEWER_H

#include "includes.h"
#include "code_tree.h"
#include "base_viewer.h"

/*
 * output viewer
 */
class output_viewer : public text_base_viewer {
  typedef text_base_viewer inherited;

private:
  vmessage *infobar;

  const char *current_file;
  code_tree *outtree;
  file_set_entry *current_fse;
  class scope_node_list *proc_scopes;

  void erase_mappings(void);
  void clear_cache(void);
  boolean build_outtree(const char *filename);
  boolean print_output(const char *filename);
  void update_infobar(void);
  void show(vnode *vn);

  /* static functions */
  static void do_open_cmd(event &e, output_viewer *viewer);

  static code_range map_tree_node(tree_node *tn, output_viewer *viewer);
  static file_set_entry *find_fse(const char *src_file);

public:
  output_viewer(void);
  ~output_viewer(void);
  virtual void create_window(void);
  virtual char *class_name(void) { return "Output Viewer"; }
  virtual void handle_event(event &e);

  /* display */
  void clear(void);
  void view(file_set_entry *fse, char *out_file);
  void view(tree_node *tn, boolean select = TRUE);

  /* state of viewer */
  code_tree *output_tree(void) { return outtree; }

  static window *constructor(void) {
    return (new output_viewer);
  }
};

#endif
